#! /bin/perl/bin -w
use strict;
use diagnostics;

# use File::DosGlob 'glob'; # lets globbing work run under PerlEz ....

# The top level code: Process() reads .qf files and writes .html;
#                     Files() returns an array of .qf files

Process(Files()) unless WebMode();

my $All = 0;

sub Files () {
  # Check for the -all cmd line switch
  $All = grep m/-all/i, @ARGV;

  # Any .qf files listed on the command line OR all .qf files in ./
  my @Result = map { m[\S+\.qf$] ? $_ : () } @ARGV;
  return @Result if @Result;
  return (<*.qf>); # if no .qf files on cmd line
}

sub Process (@) { foreach (@_) { ProcessFile($_) }}

use BasicFileIO;

sub ProcessFile {
  my $sourcename = shift;
  return unless -e $sourcename; # silently ignore (!) any non-existent files specified on cmd line

  # Return if output files exists and is not older than the input file
  (my $targetname = $sourcename) =~ s/(.*)\.qf$/$1.html/;
  return if not $All and -e $targetname and -M $sourcename > -M $targetname;

  my ($result, $error, @warnings) = GenerateQuiz($sourcename, ReadFile $sourcename);
  print +($error ? $error : join "\n", @warnings), "\n";
  unless ($error) {
    WriteFile($targetname, $result);
    print "$sourcename -> $targetname\n";
  }
}

use Template;

# Returns a three element list: ($result, $error, @warnings)
sub GenerateQuiz {
  my ($sourcename, $sourcefile) = @_;
  my ($result, @warnings) = ('', ());

  # The sourcefile may contain C++ style comments
  $sourcefile =~ s[ // [^\n]* \n ][\n]gmx;  # strip // comments
  $sourcefile =~ s[ /\* .*? \*/ ][ ]gsx;    # strip /**/ comments

  # The sourcefile consists of a series of pps, separated by 1 or more blank lines
  my ($title, $body, $questions, $responses, @scores) = ('Quiz', '', 0, 0, ());
  foreach my $paragraph (split '(?:\n\s*\n)+', $sourcefile) {
    $paragraph =~ s[\s*][]s; # strip leading white space
    my $first = substr $paragraph, 0, 1; # get the code char, if any

    if ($first eq '+') {      # The title
      $title = substr $paragraph, 1; # strip the #
      $title =~ s[\s+][ ]sg; # remove any tabs or newlines that would confound <title>
    }
    elsif ($first eq '?') { # A question
      my @pieces = split '\n\s+', $paragraph; # responses are lines starting w/ a space
      my $text = substr((shift @pieces), 1);
      $body .= join '', "\n<ol start=", ++$questions, ">\n<li><a name=Q$questions>$text</a><br>\n";
      foreach my $response (@pieces) {
        my ($text, $value) = $response =~ m[ (.*) \s* = \s* ([+-]? \d+ [.]? \d* \s* $|
                                                             [+-]? . \d+ \s* $)]x;
        $text = $response unless defined $text;
        $value = 0 unless defined $value;
        $value += 0; # force numeric, guarantee all value= values are in Perl std fmt
        $body .= join '', qq{<input type="radio" name="Q$questions" value="$value" },
                          q{onClick="AutoScroll('Q}, $questions + 1, q{')">},
                          q{<a href="javascript:Click(},
                          $responses++, q{, 'Q}, $questions + 1, qq{')">$text</a><br>\n};
      }
      $body .= "</ol>\n";
    }
    elsif ($first eq '*') { # A score range
      my ($scores, $text) = split '\n', $paragraph, 2;
      my ($low, $high) = $scores =~ m[\* \s* ([+-]? \d+) \s* - \s* ([+-]? \d+)]x;
      unless (defined $low) { ($low) = $scores =~ m[\* \s* ([+-]? \d+)]x }
      push @scores, [$text, $low, defined $high ? $high : ()];
    }
    else {                  # HTML text
      $body .= "\n\n<p>$paragraph"
    }
  }
  return ($result,
          "ERROR in $sourcename:\n\tNo questions (paragraphs starting with a ?).",
          @warnings) unless $questions;
  $questions++; $body .= "<a name=Q$questions></a>\n";

  # Convert @scores to HTML code. $scores will be JavaScript string literal.
  my ($scores, @anchors) = ("'<table>' +\n", ());
  foreach my $scoreref (@scores = sort {$a->[1] <=> $b->[1]} @scores) {
    my ($text, $low, $high) = @$scoreref;
    $text =~ s(')(\\')g;     # escape single quotes
    $text =~ s(\n)( ' +\n')sg; # JavaScript doesn't allow linebreaks in string literals
    $scores .= join '', "'<tr><td valign=top width=20%>$low",
                        defined $high ? " - $high" : '',
                        qq{' +\n'<td valign=top><a name="$low">$text</a>' + \n};
    push @anchors, "'$low'";
  }
  $scores .= "'</table>'";

  # Validate score tables
  return ($result,
          "ERROR in $sourcename:\n\tNo scores (paragraphs starting with a *).",
          @warnings) unless @anchors;
  foreach my $scoreref (@scores) { # check that low < high
    my ($text, $low, $high) = @$scoreref ;
    if (defined $high and $low >= $high)
      { return ($result,
                "ERROR in $sourcename:\n\tLow score ($low) >= high score ($high).\n",
                @warnings) }
  }
  for (my $index=0; $index < @scores - 1; $index++) {
    my ($thistext, $thislow, $thishigh) = @{ $scores[$index]   };
    my ($nexttext, $nextlow, $nexthigh) = @{ $scores[$index+1] };
    if (defined $thishigh and $thishigh >= $nextlow)
      { push @warnings, "WARNING in $sourcename:\n\tScore range $thislow - $thishigh includes next low score, $nextlow.\n" }
    $thishigh = $thislow unless defined $thishigh;
    $nexthigh = $nextlow unless defined $nexthigh;
    if ($nextlow != $thishigh + 1)
      {  push @warnings, "WARNING in $sourcename:\n\tGap between score ranges $thislow - $thishigh and $nextlow - $nexthigh.\n" }
  }

  # Do $Template substitutions
  $result = $Template;
  $result =~ s[::LONDONTIME::][&LondonTime]e;
  $result =~ s[::TITLE::][$title]g; # <title> and <h1>
  $result =~ s[::ANCHORS::][join ', ', @anchors]e;
  $result =~ s[::EXPLANATION::][$scores];
  $result =~ s[::BODY::][$body];

  # Successful return
  return ($result, '', @warnings);
}

sub LondonTime {
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime;
  return "$hour:$min:$sec GMT $mday-" .
         (qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec))[$mon] . '-'. $year % 100;
}

# Web support: WebMode() calls GenerateQuiz and returns 1 iff running as CGI

use CGI;

sub WebMode {
  # CGI detect by presense of GATEWAY_INTERFACE env car and no args,
  # ala ppalfrad@cosy.sbg.ac.at's code2html.pl
  return 0 unless defined $ENV{'GATEWAY_INTERFACE'} and @ARGV == 0;

  my $cgi= new CGI;

  my $source = $cgi->param('quizsource');
  my $file   = $cgi->param('quizfile');
  my $datatype = $source ? 'WEB PAGE INPUT' : 'UPLOADED FILE';
  if ($file and not $source) { local $/; $source = <$file> }

  $source =~ s[\r][]gs; # zap cr's
  my ($result, $error, @warnings) = GenerateQuiz($datatype, $source);

  print $cgi->header, $result unless $error;
  print $cgi->header, $cgi->start_html('error'), $error, $cgi->end_html if $error;

  return 1;
}
